/******************************************************************************
* 		                                         COPYRIGHT (c) MOTOROLA 2001   
* FILE NAME: basic.C                                                      
*                                                                           
* PURPOSE: .WAV to .C (array file) converter   											 
*                                                                           
*******************************************************************************
*******************************************************************************  
**  THIS CODE IS ONLY INTENDED AS AN EXAMPLE FOR THE METROWERKS COMPILER     **  
**      AND HAS ONLY BEEN GIVEN A MIMIMUM LEVEL OF TEST.                     **  
**  IT IS PROVIDED 'AS SEEN' WITH NO GUARANTEES AND NO PROMISE OF SUPPORT.   **  
*******************************************************************************  
*******************************************************************************
*                                                                           
* DOCUMENTATION SOURCE: EKB Apps                                                                                                                                             
*                                                                           
* MEMORY USAGE(bytes)  RAM:         ROM:                                    
*                                                                           
* INCLUDE FILES:  stdio.h,                                                
*                                                                           
* FUNCTIONS :  main                                                         
*                                                                                                                        
* COMPILER: Metrowerks                     VERSION: 6.0                     
*                                                                           
* DESCRIPTION:            
*                                                                                                                                                                      
*                                                                         
* AUTHOR: Grant More LOCATION: EKB      LAST EDIT DATE: 15.11.01 
*                                                                         
* UPDATE HISTORY                                                         
* REV      AUTHOR       DATE       DESCRIPTION OF CHANGE                  
* ---      ------      ---------    ---------------------                 
* 1.0      G More       3/11/01     initial coding began  
* 2.0	   G More		15/11/01	revised program function              
*
******************************************************************************/                                                                        


/*===========================================================================*/
/* Motorola reserves the right to make changes without further notice to any */
/* product herein to improve reliability, function, or design. Motorola does */
/* not assume any  liability arising  out  of the  application or use of any */
/* product,  circuit, or software described herein;  neither  does it convey */
/* any license under its patent rights  nor the  rights of others.  Motorola */
/* products are not designed, intended,  or authorized for use as components */
/* in  systems  intended  for  surgical  implant  into  the  body, or  other */
/* applications intended to support life, or  for any  other application  in */
/* which the failure of the Motorola product  could create a situation where */
/* personal injury or death may occur. Should Buyer purchase or use Motorola */
/* products for any such intended  or unauthorized  application, Buyer shall */
/* indemnify and  hold  Motorola  and its officers, employees, subsidiaries, */
/* affiliates,  and distributors harmless against all claims costs, damages, */
/* and expenses, and reasonable  attorney  fees arising  out of, directly or */
/* indirectly,  any claim of personal injury  or death  associated with such */
/* unintended or unauthorized use, even if such claim alleges that  Motorola */
/* was negligent regarding the  design  or manufacture of the part. Motorola */
/* and the Motorola logo* are registered trademarks of Motorola Ltd.         */
/*****************************************************************************/


/* include files */

#include <stdio.h>
#include <string.h>

/******************************************************************************
Function Name	:	main
Engineer		:	G.More
Date			:	15.11.01

Parameters		:	NONE
Returns			:	NONE
*******************************************************************************/

main(void) 
  	{
  		unsigned char convertASCII(unsigned char Nibble);
  		
  		unsigned long riffCheck, monoCheck, rateCheck, fileSize; /* Variables read from input file*/
  		unsigned char sampleCheck, upperNibble, lowerNibble; 

  		unsigned char dataArray[240000]; /*Declares 240 kB data buffer to store input .WAV file*/
  		unsigned char asciiBuffer[5]; /* Buffer used to construct ASCII data before passing to output file */
  		
      	union 	{
       	unsigned long wavSize_long; 
       	unsigned char wavSize_char[4];
       	} wavsize;
  		
  		char stringBuffer[3]; /*Intermediate buffer used for output file naming */
 		char filenameString[10]; /* As above */ 

  		int x, y, pageNumber, outfileSize; /* Internal variables */
  	  	  	  	 	  	
  		FILE *in_file; /*Initialize file for input, generate pointer to current address in file */
  		FILE *out_file; /*Initialize file for output, generate pointer to current address in file */
  		
  		in_file = fopen("input.wav","rb"); /*Open file for input*/
  		
  		fseek(in_file,0L,0); /*Point to beginning of in_file*/
  				
		/* Make sure input file is in the required format*/
		
  		fread(&riffCheck, 4, 1, in_file); /*Read "RIFF" from bytes 0-3 of in_file*/
  		
  		fseek(in_file,18L,1); /*Skip 18 bytes further into in_file*/
  		fread(&monoCheck, 4, 1, in_file); /*Read mono/stereo flag value*/
  		  		
  		fseek(in_file,2L,1); /*Skip 2 bytes further into in_file*/  		
  		fread(&rateCheck, 4, 1, in_file); /*Read sample rate*/
  		
  		fread(&sampleCheck, 2, 1, in_file); /*Read bytes per sample flag value*/
  		
		fseek(in_file,6L,1); /*Skip 6 bytes further into in_file*/  		
  		fread(&fileSize, 4, 1, in_file); /*Read file size data*/
 		
  		if ((riffCheck != 0x46464952) && (monoCheck != 0x0001) && (rateCheck != 0x2B11) && (sampleCheck != 0x0001) && (fileSize <= 240000)) /* If filetype is OK*/
  			{
  				printf("ERROR: Input file is not of 11.025kHz 8-bit Mono .WAV type, or no file detected.\n");
  				printf("ERROR: File must be smaller than 240kB.\n"); /*Notify user input file not detected or wrong .WAV type*/
   			} /*End of if*/
  		else 
  			{	
  				printf("Correct file type detected, continuing with array file creation.\n"); /*Notify user correct file detected*/  					
  				
  				fseek(in_file, 1L, 1);  /*Skip a byte */ 				
  				fread(dataArray, 1, fileSize, in_file); /*Read data at current location in in_file into data buffer */
  				
 				printf("End of input file reached, data buffered.\n"); /*Notify user read operation complete*/
 	
 				pageNumber = (int) (fileSize/16383); /*Calculates integer number of pages*/
 				pageNumber++; /*Adds one to mop up last page */
				 				
 				for (y=1; y<=pageNumber; y++) /*Do for number of pages */
 				{
 					if (y != pageNumber) /* If not the last page */
 						outfileSize = 16383; /* Page is maximum page length */
 					else	
 						outfileSize = (fileSize-(16383*(pageNumber-1)+2)); /* Page is length of remaining data */
 						
 					wavsize.wavSize_long = outfileSize;
 					
 					strcpy(filenameString, "output"); /*Name control for output file */
 					filenameString[6] = 0; /* Null to signify end of string */
 					
 					if (y <=9) /* If page number < 10*/
 					{
 						strcpy (stringBuffer, "0"); /* Write page number to stringBuffer */
 						stringBuffer[1] = convertASCII(y);
						stringBuffer[2] = 0; /* Null to give end of srting */
 					}		
 					else /* If page number > 10*/
 					{
 						strcpy (stringBuffer, "1"); /* Write page number to stringBuffer */
 						stringBuffer[1] = (convertASCII(-(10-y)));
 						stringBuffer[2] = 0;
 					}
 					strcat(filenameString, stringBuffer); /* Write stringBuffer to filenameString */
 					strcat(filenameString, ".c"); /* Add file extension */
 										 									 				
					out_file = fopen(filenameString,"wb"); /*Open file for output*/
					fseek(out_file,0L,0); /*Point to beginning of out_file*/
 				
 					fwrite("#pragma CONST_SEG PPAGE WAV", 1, 27, out_file); /* Write C header into out_file*/
 					 					 					
 					fwrite(stringBuffer, 1, 2, out_file); /* C header*/
 					fwrite("\n", 1, 1, out_file);
 					
 					fwrite("const unsigned char wav", 1, 23, out_file); /*C header*/
 					fwrite(stringBuffer, 1, 2, out_file); 
 					fwrite("[]={", 1, 4, out_file);
 					
	 				for (x=0; x<=outfileSize; x++) /* Do this for every byte of bata in dataBuffer*/
 					{	
 						if (x%10 == 0) /* If x is divisible by 10*/
 						{
 							fwrite("\n", 1, 1, out_file); /* Write a newline character (makes output readable) */
 						}
 					 	
 					 	if (x<=1)
	 					{
	 					 	upperNibble = (((wavsize.wavSize_char[1-x]) & 0xF0)>>4); /* Split data into nibbles for ASCII conversion */
 							lowerNibble = ((wavsize.wavSize_char[1-x]) & 0x0F);
 					 	}
 					 	else
 					 	{				
 							upperNibble = ((dataArray[(x + (y-1)*16383)] & 0xF0)>>4); /* Split data into nibbles for ASCII conversion */
 							lowerNibble = (dataArray[(x + (y-1)*16383)] & 0x0F);
 						}
 					 					    
 						asciiBuffer[0] = 0x30; /* ASCII "0" */
 						asciiBuffer[1] = 'x';
 						asciiBuffer[2] = convertASCII(upperNibble); /*Writes ASCII for upperNibble to asciiBuffer*/
 						asciiBuffer[3] = convertASCII(lowerNibble); /*Same as above but lowerNibble */
 						if (x<=(outfileSize-1)) /*If not end of page */
 							asciiBuffer[4] = ',';
 						else
 							asciiBuffer[4] = '}'; /* If end of page */
 						fwrite(asciiBuffer, 1, 5, out_file); /*Write buffer to file */
 					}	
 				
 					fwrite(";", 1, 1, out_file); /* End the file */
 					printf("Output file written.\n"); /*Inform user output file written OK*/
 					fclose(out_file); /*Close out_file*/
				}	 			
 			fclose(in_file); /*Close in_file*/
 
 			printf("Task complete, all good.\n"); /*Inform user task complete*/
 			} /* End of else*/

} /*End of program*/

/******************************************************************************
Function Name	:	convertASCII
Engineer		:	G.More
Date			:	15.11.01

Parameters		:	Nibble of binary data (0000 to 1111) (unsigned char)
Returns			:	ASCII code for nibble (unsigned char)
*******************************************************************************/

	unsigned char convertASCII(unsigned char Nibble)
	{	
		if (Nibble <= 0x09) /* If nibble is less than 9 (ie. a number) */
			return (Nibble + 0x30); /* Return the ASCII code for this number */
		else
			return (Nibble + 0x37); /* Return the ASCII code for this letter */
	} 














	